@@ -1,33 +1,30 @@
 #include <stdio.h>
 #include <string.h>
+#include <stdlib.h>
 
 #include "different.h"
 #include "gripes.h"
 #include "man-config.h"
 #include "util.h"
 
-#define BUFSIZE 4096
-
 static struct filelist {
     char *pathname;
     struct filelist *next;
 } cat_list, man_list;
 
 static int
-is_different(char *file, struct filelist *p) {
-    char command[BUFSIZE];
-    char *cmp = getval("CMP");
+is_different(const char *file, struct filelist *p) {
+    char *command;
+    const char *cmp = getval("CMP");
     int ret;
 
     if (cmp) {
 	while (p->next) {
-	    if (strlen(cmp) + strlen (file) + strlen(p->pathname) + 4 < BUFSIZE) {
-		sprintf(command, "%s %s %s\n", cmp, file, p->pathname);
-		ret = do_system_command (command, 1);
-		if (ret == 0) {
-		    gripe(IDENTICAL, file, p->pathname);
-		    return 0;
-		}
+	    command = my_xsprintf("%s %S %S\n", cmp, file, p->pathname);
+	    ret = do_system_command (command, 1);
+	    if (ret == 0) {
+		gripe(IDENTICAL, file, p->pathname);
+		return 0;
 	    }
 	    p = p->next;
 	}
@@ -39,11 +36,11 @@ is_different(char *file, struct filelist *p) {
 }
 
 int
-different_cat_file (char *file) {
+different_cat_file (const char *file) {
     return is_different (file, &cat_list);
 }
 
 int
-different_man_file (char *file) {
+different_man_file (const char *file) {
     return is_different (file, &man_list);
 }
