@@ -35,6 +35,7 @@
 #define R_OK 4
 #endif
 
+extern char *index (const char *, int);		/* not always in <string.h> */
 extern char *rindex (const char *, int);	/* not always in <string.h> */
 
 #include "defs.h"
@@ -44,27 +45,43 @@ extern char *rindex (const char *, int);	/* not always in <string.h> */
 #include "manpath.h"
 #include "man-config.h"
 #include "man-getopt.h"
+#include "man-iconv.h"
 #include "to_cat.h"
 #include "util.h"
 #include "glob.h"
 #include "different.h"
+#include "man-iconv.h"
 
-char *progname;
-char *pager;
+#define SIZE(x) (sizeof(x)/sizeof((x)[0]))
+
+const char *progname;
+const char *pager, *browser, *htmlpager;
 char *colon_sep_section_list;
 char *roff_directive;
 char *dohp = 0;
+int do_irix;
+int do_win32;
 int apropos;
 int whatis;
-int nocats;
+int nocats;			/* set by -c option: do not use cat page */
+				/* this means that cat pages must not be used,
+				   perhaps because the user knows they are
+				   old or corrupt or so */
+int can_use_cache;		/* output device is a tty, width 80 */
+				/* this means that the result may be written
+				   in /var/cache, and may be read from there */
 int findall;
 int print_where;
 int one_per_line;
 int do_troff;
+int preformat;
 int debug;
 int fhs;
 int fsstnd;
-
+int noautopath;
+int nocache;
+static int is_japanese;
+static char *language;
 static char **section_list;
 
 #ifdef DO_COMPRESS
@@ -73,11 +90,11 @@ int do_compress = 1;
 int do_compress = 0;
 #endif
 
-#define BUFSIZE 4096
+#define BUFSIZE 8192
 
 /*
  * Try to determine the line length to use.
- * Preferences: 1. MANWIDTH 2. ioctl 3. COLUMNS 4. 80
+ * Preferences: 1. MANWIDTH, 2. ioctl, 3. COLUMNS, 4. 80
  *
  * joey, 950902
  */
@@ -92,6 +109,10 @@ get_line_length(void){
      char *cp;
      int width;
 
+     if (preformat) {
+	  line_length = 80;
+	  return;
+     }
      if ((cp = getenv ("MANWIDTH")) != NULL && (width = atoi(cp)) > 0) {
 	  line_length = width;
 	  return;
@@ -121,6 +142,28 @@ setll(void) {
 	  line_length*9/10 : 0;
 }
 
+/* People prefer no page headings in their man screen output;
+   now ".pl 0" has a bad effect on .SH etc, so we need ".pl N"
+   for some large number N, like 1100i (a hundred pages). */
+#define VERY_LONG_PAGE	"1100i"
+
+static char *
+setpl(void) {
+     char *pl;
+     if (do_troff)
+	  return NULL;
+     if (preformat)
+	  pl = VERY_LONG_PAGE;
+     else
+     if ((pl = getenv("MANPL")) == 0) {
+	  if (isatty(0) && isatty(1))
+	       pl = VERY_LONG_PAGE;
+	  else
+	       pl = "11i";		/* old troff default */
+     }
+     return pl;
+}
+
 /*
  * Check to see if the argument is a valid section number.  If the
  * first character of name is a numeral, or the name matches one of
@@ -129,7 +172,7 @@ setll(void) {
  * named directories like .../man3f.  Yuk. 
  */
 static char *
-is_section (char * name) {
+is_section (char *name) {
      char **vs;
 
      /* 3Xt may be a section, but 3DBorder is a man page */
@@ -159,13 +202,13 @@ remove_file (char *file) {
 }
 
 static void
-remove_other_catfiles (char *catfile) {
-     char pathname[BUFSIZE];
+remove_other_catfiles (const char *catfile) {
+     char *pathname;
      char *t;
      char **gf;
      int offset;
 
-     strcpy (pathname, catfile);
+     pathname = my_strdup(catfile);
      t = rindex(pathname, '.');
      if (t == NULL || strcmp(t, getval("COMPRESS_EXT")))
 	  return;
@@ -199,26 +242,29 @@ remove_other_catfiles (char *catfile) {
  * Simply display the preformatted page.
  */
 static int
-display_cat_file (char *file) {
+display_cat_file (const char *file) {
      int found;
-     char command[BUFSIZE];
+
+     if (preformat)
+	  return 1;		/* nothing to do - preformat only */
 
      found = 0;
 
      if (access (file, R_OK) == 0 && different_cat_file(file)) {
-	  char *expander = get_expander (file);
+	  char *command = NULL;
+	  const char *expander = get_expander (file);
 
 	  if (expander != NULL && expander[0] != 0) {
 	       if (isatty(1))
-		    sprintf (command, "%s %s | %s", expander, file, pager);
+		    command = my_xsprintf("%s %S | %s", expander, file, pager);
 	       else
-		    sprintf (command, "%s %s", expander, file);
+		    command = my_xsprintf("%s %S", expander, file);
 	  } else {
-	       if (isatty(1))
-		    sprintf (command, "%s %s", pager, file);
-	       else {
-		    char *cat = getval("CAT");
-		    sprintf (command, "%s %s", cat ? cat : "cat", file);
+	       if (isatty(1)) {
+		    command = my_xsprintf("%s %S", pager, file);
+	       } else {
+		    const char *cat = getval("CAT");
+		    command = my_xsprintf("%s %S", cat[0] ? cat : "cat", file);
 	       }
 	  }
 	  found = !do_system_command (command, 0);
@@ -226,6 +272,30 @@ display_cat_file (char *file) {
      return found;
 }
 
+/*
+ * Simply display the preformatted page.
+ */
+static int
+display_html_file (const char *file) {
+     int found;
+
+     found = 0;
+
+     if (access (file, R_OK) == 0 && different_cat_file(file)) {
+	  char *command = NULL;
+
+	  if (isatty(1)) {
+	       command = my_xsprintf("%s %S", browser, file);
+	  } else {
+	       command = my_xsprintf("%s %S", htmlpager, file);
+	  }
+	  found = !do_system_command (command, 0);
+     }
+     return found;
+
+     return 1;
+}
+
 /*
  * Try to find the ultimate source file.  If the first line of the
  * current file is not of the form
@@ -242,11 +312,11 @@ display_cat_file (char *file) {
  * (which contains `.so TIFFsize.3t')
  * we return /usr/man/man3/TIFFsize.3t .
  */
-static char *
-ultimate_source (char *name0) {
+static const char *
+ultimate_source (const char *name0) {
      FILE *fp;
      char *name;
-     char *expander;
+     const char *expander;
      int expfl = 0;
      char *fgr;
      char *beg;
@@ -255,29 +325,32 @@ ultimate_source (char *name0) {
      char buf[BUFSIZE];
      static char ultname[BUFSIZE];
 
-     name = my_strdup(name0);
+     if (strlen(name0) >= sizeof(ultname))
+	     return name0;
+     strcpy(ultname, name0);
+     name = ultname;
 
 again:
      expander = get_expander (name);
      if (expander && *expander) {
-	  char command[BUFSIZE];
+	  char *command;
 
-	  sprintf (command, "%s %s", expander, name);
+	  command = my_xsprintf ("%s %S", expander, name);
 	  fp = my_popen (command, "r");
 	  if (fp == NULL) {
 	       perror("popen");
 	       gripe (EXPANSION_FAILED, command);
 	       return (NULL);
 	  }
-	  fgr = fgets (buf, BUFSIZE, fp);
+	  fgr = fgets (buf, sizeof(buf), fp);
 	  pclose (fp);
 	  expfl = 1;
      } else {
 	  fp = fopen (name, "r");
 	  if (fp == NULL && expfl) {
 	       char *extp = rindex (name0, '.');
-	       if (extp && *extp) {
-		    strcat(name, extp);		/* name==ultname here */
+	       if (extp && *extp && strlen(name)+strlen(extp) < BUFSIZE) {
+		    strcat(name, extp);
 		    fp = fopen (name, "r");
 	       }
 	  }
@@ -286,7 +359,8 @@ again:
 	   * .so files - we could glob for all possible extensions,
 	   * for now: only try .gz
 	   */
-	  else if (fp == NULL && get_expander(".gz")) {
+	  else if (fp == NULL && get_expander(".gz") &&
+		   strlen(name)+strlen(".gz") < BUFSIZE) {
 	       strcat(name, ".gz");
 	       fp = fopen (name, "r");
 	  }
@@ -296,7 +370,7 @@ again:
 	       gripe (OPEN_ERROR, name);
 	       return (NULL);
 	  }
-	  fgr = fgets (buf, BUFSIZE, fp);
+	  fgr = fgets (buf, sizeof(buf), fp);
 	  fclose (fp);
      }
 
@@ -307,7 +381,7 @@ again:
      }
 
      if (strncmp(buf, ".so", 3))
-	  return (name);
+	  return (my_strdup(name));
 
      beg = buf+3;
      while (*beg == ' ' || *beg == '\t')
@@ -315,30 +389,33 @@ again:
 
      end = beg;
      while (*end != ' ' && *end != '\t' && *end != '\n' && *end != '\0')
-	  end++;
-
+	  end++;		/* note that buf is NUL-terminated */
      *end = '\0';
 
-     if (name != ultname) {
-	  strcpy(ultname, name);
-	  name = ultname;
-     }
-
      /* If name ends in path/manx/foo.9x then use path, otherwise
 	try same directory. */
      if ((cp = rindex(name, '/')) == NULL) /* very strange ... */
 	  return 0;
      *cp = 0;
+
+     /* allow "man ./foo.3" where foo.3 contains ".so man2/bar.2" */
+     if ((cp = rindex(name, '/')) != NULL && !strcmp(cp+1, "."))
+	  *cp = 0;
+
+     /* In all cases, the new name will be something from name
+	followed by something from beg. */
+     if (strlen(name) + strlen(beg) + 1 >= BUFSIZ)
+	  return 0;		/* very long names, ignore */
+
      if (!index(beg, '/')) {
 	  /* strange.. try same directory as the .so file */
 	  strcat(name, "/");
 	  strcat(name, beg);
-     } else
-     if((cp = rindex(name, '/')) != NULL && !strncmp(cp+1, "man", 3))
+     } else if((cp = rindex(name, '/')) != NULL && !strncmp(cp+1, "man", 3)) {
 	  strcpy(cp+1, beg);
-     else if((cp = rindex(beg, '/')) != NULL)
+     } else if((cp = rindex(beg, '/')) != NULL) {
 	  strcat(name, cp);
-     else {
+     } else {
 	  strcat(name, "/");
 	  strcat(name, beg);
      }
@@ -347,13 +424,17 @@ again:
 }
 
 static void
-add_directive (char *d, char *file, char *buf) {
+add_directive (const char *d, const char *file, char *buf, int buflen) {
      if ((d = getval(d)) != 0 && *d) {
 	  if (*buf == 0) {
+	       if (strlen(d) + strlen(file) + 2 > buflen)
+		    return;
 	       strcpy (buf, d);
 	       strcat (buf, " ");
 	       strcat (buf, file);
 	  } else {
+	       if (strlen(d) + strlen(buf) + 4 > buflen)
+		    return;
 	       strcat (buf, " | ");
 	       strcat (buf, d);
 	  }
@@ -361,47 +442,68 @@ add_directive (char *d, char *file, char *buf) {
 }
 
 static int
-parse_roff_directive (char *cp, char *file, char *buf) {
+is_lang_page (char *lang, const char *file) {
+	char lang_path[16] = "";
+
+	snprintf(lang_path, sizeof(lang_path), "/%s/", lang);
+	if (strstr(file, lang_path))
+		return 1;
+	if (strlen(lang) > 2) {
+		lang_path[3] = '/';
+		lang_path[4] = 0;
+		if (strstr(file, lang_path))
+			return 1;
+	}
+	return 0;
+}
+
+static int
+parse_roff_directive (char *cp, const char *file, char *buf, int buflen) {
      char c;
      int tbl_found = 0;
+     int use_jroff;
+
+     use_jroff = (is_japanese &&
+		   (strstr(file, "/jman/") || is_lang_page(language, file)));
 
      while ((c = *cp++) != '\0') {
 	  switch (c) {
 	  case 'e':
 	       if (debug)
 		    gripe (FOUND_EQN);
-	       add_directive ((do_troff ? "EQN" : "NEQN"), file, buf);
+	       add_directive((do_troff ? "EQN" : use_jroff ? "JNEQN": "NEQN"),
+			     file, buf, buflen);
 	       break;
 
 	  case 'g':
 	       if (debug)
 		    gripe (FOUND_GRAP);
-	       add_directive ("GRAP", file, buf);
+	       add_directive ("GRAP", file, buf, buflen);
 	       break;
 
 	  case 'p':
 	       if (debug)
 		    gripe (FOUND_PIC);
-	       add_directive ("PIC", file, buf);
+	       add_directive ("PIC", file, buf, buflen);
 	       break;
 
 	  case 't':
 	       if (debug)
 		    gripe (FOUND_TBL);
 	       tbl_found++;
-	       add_directive ("TBL", file, buf);
+	       add_directive ("TBL", file, buf, buflen);
 	       break;
 
 	  case 'v':
 	       if (debug)
 		    gripe (FOUND_VGRIND);
-	       add_directive ("VGRIND", file, buf);
+	       add_directive ("VGRIND", file, buf, buflen);
 	       break;
 
 	  case 'r':
 	       if (debug)
 		    gripe (FOUND_REFER);
-	       add_directive ("REFER", file, buf);
+	       add_directive ("REFER", file, buf, buflen);
 	       break;
 
 	  case ' ':
@@ -418,49 +520,118 @@ done:
      if (*buf == 0)
 	  return 1;
 
-     strcat (buf, " | ");
-     strcat (buf, getval(do_troff ? "TROFF" : "NROFF"));
+     add_directive (do_troff ? "TROFF" : use_jroff ? "JNROFF" : "NROFF",
+		    "", buf, buflen);
 
-     if (tbl_found && !do_troff && *getval("COL")) {
-	  strcat (buf, " | ");
-	  strcat (buf, getval("COL"));
-     }
+     if (tbl_found && !do_troff && *getval("COL"))
+	  add_directive ("COL", "", buf, buflen);
 
      return 0;
 }
 
 static char *
-make_roff_command (char *file) {
+eos(char *s) {
+     while(*s) s++;
+     return s;
+}
+
+/*
+ * Create command to format FILE, in the directory PATH/manX
+ */
+static char *
+make_roff_command (const char *path, const char *file) {
      FILE *fp;
-     char line [BUFSIZE];
      static char buf [BUFSIZE];
+     char line [BUFSIZE], bufh [BUFSIZE], buft [BUFSIZE];
      int status, ll;
-     char *cp, *expander, *fgr;
+     char *cp, *fgr, *pl;
+     char *command = "";
+     const char *expander;
+     const char *converter;
 
      /* if window size differs much from 80, try to adapt */
-     /* (but write only standard formatted files to the cat directory) */
+     /* (but write only standard formatted files to the cat directory,
+	see can_use_cache) */
      ll = setll();
+     pl = setpl();
+     if (ll && debug)
+	  gripe (NO_CAT_FOR_NONSTD_LL);
 
      expander = get_expander (file);
-     if (expander && *expander) {
-	  if (ll)
-	       sprintf (buf, "(echo \".ll %d.%di\"; %s %s)",
-			ll/10, ll%10, expander, file);
-	  else
-	       sprintf (buf, "%s %s", expander, file);
-     } else {
-	  if (ll)
-	       sprintf (buf, "(echo \".ll %d.%di\"; cat %s)",
-			ll/10, ll%10, file);
-	  else
-	       buf[0] = 0;
+     converter = get_converter (path);
+
+     /* head */
+     bufh[0] = 0;
+     if (ll || pl) {
+	  /* some versions of echo do not accept the -e flag,
+	     so we just use two echo calls when needed */
+	  strcat(bufh, "(");
+	  if (ll) {
+	       /*
+		* We should set line length and title line length.
+		* However, a .lt command here fails, only
+		*  .ev 1; .lt ...; .ev helps for my version of groff.
+		* The LL assignment is needed by the mandoc macros.
+		*/
+	       sprintf(eos(bufh), "echo \".ll %d.%di\"; ", ll/10, ll%10);
+	       sprintf(eos(bufh), "echo \".nr LL %d.%di\"; ", ll/10, ll%10);
+#if 0
+	       sprintf(eos(bufh), "echo \".lt %d.%di\"; ", ll/10, ll%10);
+#endif
+	  }
+	  if (pl)
+	       sprintf(eos(bufh), "echo \".pl %.128s\"; ", pl);
      }
 
+     /* tail */
+     buft[0] = 0;
+     if (ll || pl) {
+	  if (pl && !strcmp(pl, VERY_LONG_PAGE))
+	      /* At end of the nroff source, set the page length to
+		 the current position plus 10 lines.  This plus setpl()
+		 gives us a single page that just contains the whole
+		 man page. (William Webber, wew@cs.rmit.edu.au) */
+	      strcat(buft, "; echo \".\\\\\\\"\"; echo \".pl \\n(nlu+10\"");
+#if 0
+	      /* In case this doesnt work for some reason,
+		 michaelkjohnson suggests: I've got a simple
+		 awk invocation that I throw into the pipeline: */
+
+		 awk 'BEGIN {RS="\n\n\n\n*"} /.*/ {print}'
+#endif
+	  strcat(buft, ")");
+     }
+
+     if (expander && *expander) {
+	  if (converter && *converter)
+	     command = my_xsprintf("%s%s '%S' | %s%s",
+				   bufh, expander, file, converter, buft);
+	  else
+	     command = my_xsprintf("%s%s '%S'%s",
+				   bufh, expander, file, buft);
+     } else if (ll || pl) {
+	  const char *cat = getval("CAT");
+	  if (!cat || !*cat)
+		  cat = "cat";
+
+	  if (converter && *converter)
+	      command = my_xsprintf("%s%s '%S' | %s%s",
+				    bufh, cat, file, converter, buft);
+	  else
+	      command = my_xsprintf("%s%s '%S'%s",
+				    bufh, cat, file, buft);
+     }
+
+     if (strlen(command) >= sizeof(buf))
+	  exit(1);
+     strcpy(buf, command);
+
      if (roff_directive != NULL) {
 	  if (debug)
 	       gripe (ROFF_FROM_COMMAND_LINE);
 
-	  status = parse_roff_directive (roff_directive, file, buf);
+	  status = parse_roff_directive (roff_directive, file,
+					 buf, sizeof(buf));
 
 	  if (status == 0)
 	       return buf;
@@ -470,16 +641,14 @@ make_roff_command (char *file) {
      }
 
      if (expander && *expander) {
-	  char command[BUFSIZE];
-
-	  sprintf (command, "%s %s", expander, file);
-	  fp = my_popen (command, "r");
+	  char *cmd = my_xsprintf ("%s %S", expander, file);
+	  fp = my_popen (cmd, "r");
 	  if (fp == NULL) {
 	       perror("popen");
-	       gripe (EXPANSION_FAILED, command);
+	       gripe (EXPANSION_FAILED, cmd);
 	       return (NULL);
 	  }
-	  fgr = fgets (line, BUFSIZE, fp);
+	  fgr = fgets (line, sizeof(line), fp);
 	  pclose (fp);
      } else {
 	  fp = fopen (file, "r");
@@ -488,7 +657,7 @@ make_roff_command (char *file) {
 	       gripe (OPEN_ERROR, file);
 	       return (NULL);
 	  }
-	  fgr = fgets (line, BUFSIZE, fp);
+	  fgr = fgets (line, sizeof(line), fp);
 	  fclose (fp);
      }
 
@@ -503,7 +672,7 @@ make_roff_command (char *file) {
 	  if (debug)
 	       gripe (ROFF_FROM_FILE, file);
 
-	  status = parse_roff_directive (cp, file, buf);
+	  status = parse_roff_directive (cp, file, buf, sizeof(buf));
 
 	  if (status == 0)
 	       return buf;
@@ -516,7 +685,7 @@ make_roff_command (char *file) {
 	  if (debug)
 	       gripe (ROFF_FROM_ENV);
 
-	  status = parse_roff_directive (cp, file, buf);
+	  status = parse_roff_directive (cp, file, buf, sizeof(buf));
 
 	  if (status == 0)
 	       return buf;
@@ -528,7 +697,7 @@ make_roff_command (char *file) {
      if (debug)
 	  gripe (USING_DEFAULT);
 
-     (void) parse_roff_directive ("t", file, buf);
+     (void) parse_roff_directive ("t", file, buf, sizeof(buf));
 
      return buf;
 }
@@ -538,11 +707,21 @@ make_roff_command (char *file) {
  * 1 for success and 0 for failure.
  */
 static int
-make_cat_file (char *path, char *man_file, char *cat_file) {
+make_cat_file (const char *path, const char *man_file, const char *cat_file) {
      int mode;
      FILE *fp;
      char *roff_command;
-     char command[BUFSIZE];
+     char *command = NULL;
+     struct stat statbuf;
+
+     /* _Before_ first, make sure we will write to a regular file. */
+     if (stat(cat_file, &statbuf) == 0) {
+	  if(!S_ISREG(statbuf.st_mode)) {
+	       if (debug)
+		    gripe (CAT_OPEN_ERROR, cat_file);
+	       return 0;
+	  }
+     }
 
      /* First make sure we can write the file; create an empty file. */
      /* If we are suid it must get mode 0666. */
@@ -561,7 +740,7 @@ make_cat_file (char *path, char *man_file, char *cat_file) {
 
 	       /* maybe the real user can write it */
 	       /* note: just doing "> %s" gives the wrong exit status */
-	       sprintf(command, "cp /dev/null %s 2>/dev/null", cat_file);
+	       command = my_xsprintf("cp /dev/null %S 2>/dev/null", cat_file);
 	       if (do_system_command(command, 1)) {
 		    if (debug)
 			 gripe (USER_CANNOT_OPEN_CAT);
@@ -581,7 +760,7 @@ make_cat_file (char *path, char *man_file, char *cat_file) {
 		    /* probably we are sgid but not owner;
 		       just delete the file and create it again */
 		    if(unlink(cat_file) != 0) {
-			 sprintf (command, "rm %s", cat_file);
+			 command = my_xsprintf("rm %S", cat_file);
 			 (void) do_system_command (command, 1);
 		    }
 		    if ((fp = fopen (cat_file, "w")) != NULL)
@@ -590,19 +769,19 @@ make_cat_file (char *path, char *man_file, char *cat_file) {
           }
      }
 
-     roff_command = make_roff_command (man_file);
+     roff_command = make_roff_command (path, man_file);
      if (roff_command == NULL)
 	  return 0;
-     else if (do_compress)
+     if (do_compress)
 	  /* The cd is necessary, because of .so commands,
 	     like .so man1/bash.1 in bash_builtins.1.
 	     But it changes the meaning of man_file and cat_file,
 	     if these are not absolute. */
 	
-	  sprintf (command, "(cd %s ; %s | %s > %s)", path,
+	  command = my_xsprintf("(cd %S && %s | %S > %S)", path,
 		   roff_command, getval("COMPRESS"), cat_file);
      else
-	  sprintf (command, "(cd %s ; %s > %s)", path,
+	  command = my_xsprintf ("(cd %S && %s > %S)", path,
 		   roff_command, cat_file);
 
      /*
@@ -617,7 +796,7 @@ make_cat_file (char *path, char *man_file, char *cat_file) {
 	  /* success */
 	  mode = ((ruid != euid) ? 0644 : (rgid != egid) ? 0464 : 0444);
 	  if(chmod (cat_file, mode) != 0 && suid) {
-	       sprintf (command, "chmod 0%o %s", mode, cat_file);
+	       command = my_xsprintf ("chmod 0%o %S", mode, cat_file);
 	       (void) do_system_command (command, 1);
 	  }
 	  /* be silent about the success of chmod - it is not important */
@@ -626,7 +805,7 @@ make_cat_file (char *path, char *man_file, char *cat_file) {
      } else {
 	  /* something went wrong - remove garbage */
 	  if(unlink(cat_file) != 0 && suid) {
-	       sprintf (command, "rm %s", cat_file);
+	       command = my_xsprintf ("rm %S", cat_file);
 	       (void) do_system_command (command, 1);
 	  }
      }
@@ -637,19 +816,19 @@ make_cat_file (char *path, char *man_file, char *cat_file) {
 }
 
 static int
-display_man_file(char *path, char *man_file) {
+display_man_file(const char *path, const char *man_file) {
      char *roff_command;
-     char command[BUFSIZE];
+     char *command;
 
      if (!different_man_file (man_file))
 	  return 0;
-     roff_command = make_roff_command (man_file);
+     roff_command = make_roff_command (path, man_file);
      if (roff_command == NULL)
 	  return 0;
-     else if (do_troff)
-	  sprintf (command, "(cd %s ; %s)", path, roff_command);
+     if (do_troff)
+	  command = my_xsprintf ("(cd %S && %s)", path, roff_command);
      else
-	  sprintf (command, "(cd %s ; %s | %s)", path,
+	  command = my_xsprintf ("(cd %S && %s | %s)", path,
 		   roff_command, pager);
 
      return !do_system_command (command, 0);
@@ -659,16 +838,17 @@ display_man_file(char *path, char *man_file) {
  * make and display the cat file - return 0 if something went wrong
  */
 static int
-make_and_display_cat_file (char *path, char *man_file) {
-     char *cat_file;
-     char *ext;
+make_and_display_cat_file (const char *path, const char *man_file) {
+     const char *cat_file;
+     const char *ext;
      int status;
      int standards;
 
      ext = (do_compress ? getval("COMPRESS_EXT") : 0);
-     standards = (fhs ? FHS : 0) | (fsstnd ? FSSTND : 0);
 
-     if ((cat_file = convert_to_cat (man_file,ext,standards)) == NULL)
+     standards = (fhs ? FHS : 0) | (fsstnd ? FSSTND : 0) | (dohp ? DO_HP : 0);
+
+     if ((cat_file = convert_to_cat(man_file, ext, standards)) == NULL)
 	  return 0;
 
      if (debug)
@@ -680,7 +860,7 @@ make_and_display_cat_file (char *path, char *man_file) {
       * old .Z files that should be removed).
       */
 
-     status = (nocats ? -2 : is_newer (man_file, cat_file));
+     status = ((nocats | preformat) ? -2 : is_newer (man_file, cat_file));
      if (debug)
 	  gripe (IS_NEWER_RESULT, status);
      if (status == -1 || status == -3) {
@@ -733,10 +913,8 @@ make_and_display_cat_file (char *path, char *man_file) {
  * it directly.
  */
 static int
-format_and_display (char *man_file) {
-     char *path;
-     char *roff_command;
-     char command[BUFSIZE];
+format_and_display (const char *man_file) {
+     const char *path;
 
      if (access (man_file, R_OK) != 0)
 	  return 0;
@@ -753,19 +931,17 @@ format_and_display (char *man_file) {
 	  return 0;
 
      if (do_troff) {
-	  roff_command = make_roff_command (man_file);
+	  char *command;
+	  char *roff_command = make_roff_command (path, man_file);
+
 	  if (roff_command == NULL)
 	       return 0;
-	  else
-	       sprintf (command, "(cd %s ; %s)", path, roff_command);
 
+	  command = my_xsprintf("(cd %S && %s)", path, roff_command);
 	  return !do_system_command (command, 0);
      }
 
-     if (setll()) {
-	  if (debug)
-	       gripe(NO_CAT_FOR_NONSTD_LL);
-     } else if (make_and_display_cat_file (path, man_file))
+     if (can_use_cache && make_and_display_cat_file (path, man_file))
 	  return 1;
 
      /* line length was wrong or could not display cat_file */
@@ -795,7 +971,7 @@ format_and_display (char *man_file) {
  *
  */
 static int
-man (char *name, char *section) {
+man (const char *name, const char *section) {
      int found, type, flags;
      struct manpage *mp;
 
@@ -847,23 +1023,42 @@ man (char *name, char *section) {
      fflush (stdout);
      init_manpath();
 
+     can_use_cache = nocache ? 0 : (preformat || print_where ||
+		      (isatty(0) && isatty(1) && !setll()));
+
      if (do_troff) {
-	  char *t = getval("TROFF");
+	  const char *t = getval("TROFF");
 	  if (!t || !*t)
 	       return 0;	/* don't know how to format */
 	  type = TYPE_MAN;
      } else {
-	  char *n = getval("NROFF");
-	  type = TYPE_CAT;
+	  const char *n = getval("NROFF");
+	  type = 0;
+	  if (can_use_cache)
+	       type |= TYPE_CAT;
 	  if (n && *n)
 	       type |= TYPE_MAN;
 	  if (fhs || fsstnd)
 	       type |= TYPE_SCAT;
+
+	  n = getval("BROWSER");
+	  if (n && *n)
+	      type |= TYPE_HTML;
      }
 
      flags = type;
      if (!findall)
 	  flags |= ONLY_ONE;
+     if (fsstnd)
+	  flags |= FSSTND;
+     else if (fhs)
+	  flags |= FHS;
+     if (dohp)
+	  flags |= DO_HP;
+     if (do_irix)
+	  flags |= DO_IRIX;
+     if (do_win32)
+	  flags |= DO_WIN32;
 
      mp = manfile(name, section, flags, section_list, mandirlist,
 		  convert_to_cat);
@@ -877,6 +1072,12 @@ man (char *name, char *section) {
                     found = 1;
                } else
 	            found = display_cat_file(mp->filename);
+	  } else if (mp->type == TYPE_HTML) {
+               if (print_where) {
+                    printf ("%s\n", mp->filename);
+                    found = 1;
+               } else
+	            found = display_html_file(mp->filename);
 	  } else
 	       /* internal error */
 	       break;
@@ -890,7 +1091,7 @@ man (char *name, char *section) {
 static char **
 get_section_list (void) {
      int i;
-     char *p;
+     const char *p;
      char *end;
      static char *tmp_section_list[100];
 
@@ -907,7 +1108,7 @@ get_section_list (void) {
 
 	  tmp_section_list[i++] = my_strdup (p);
 
-	  if (end == NULL || i+1 == sizeof(tmp_section_list))
+	  if (end == NULL || i+1 == SIZE(tmp_section_list))
 	       break;
      }
 
@@ -915,36 +1116,38 @@ get_section_list (void) {
      return tmp_section_list;
 }
 
-static void
+/* return 0 when all was OK */
+static int
 do_global_apropos (char *name, char *section) {
      char **dp, **gf;
-     char pathname[BUFSIZE];
-     char command[BUFSIZE];
-     int res;
-
-     /* do_global_apropos() produces a long steam of `system' commands,
-	and during the system() call SIGINT and SIGQUIT are being ignored,
-	so `man -K' is difficult to interrupt.
-	However, ^Z still works, and can be followed by `kill %1'. */
+     char *pathname;
+     char *command;
+     int status, res;
 
+     status = 0;
      init_manpath();
      if (mandirlist)
 	for (dp = mandirlist; *dp; dp++) {
 	  if (debug)
 	       gripe(SEARCHING, *dp);
-	  sprintf(pathname, "%s/man%s/*", *dp, section ? section : "*");
+	  pathname = my_xsprintf("%s/man%s/*", *dp, section ? section : "*");
 	  gf = glob_filename (pathname);
+	  free(pathname);
 
 	  if (gf != (char **) -1 && gf != NULL) {
 	       for( ; *gf; gf++) {
-		    char *expander = get_expander (*gf);
+		    const char *expander = get_expander (*gf);
 		    if (expander)
-			 sprintf(command, "%s %s | grep -%c '%s'",
-				 expander, *gf, GREPSILENT, name);
+			 command = my_xsprintf("%s %S | grep '%Q'"
+					       "> /dev/null 2> /dev/null",
+				 expander, *gf, name);
 		    else
-			 sprintf(command, "grep -%c '%s' %s",
-				 GREPSILENT, name, *gf);
+			 command = my_xsprintf("grep '%Q' %S"
+					       "> /dev/null 2> /dev/null",
+				 name, *gf);
 		    res = do_system_command (command, 1);
+		    status |= res;
+		    free (command);
 		    if (res == 0) {
 			 if (print_where)
 			      printf("%s\n", *gf);
@@ -975,36 +1178,45 @@ do_global_apropos (char *name, char *section) {
 	       }
 	  }
      }
+     return status;
+}
+
+/* Special code for Japanese (to pick jnroff instead of nroff, etc.) */
+static void
+setlang(void) {
+	char *lang;
+
+	/* We use getenv() instead of setlocale(), because of
+	   glibc 2.1.x security policy for SetUID/SetGID binary. */
+	if ((lang = getenv("LANG")) == NULL &&
+	    (lang = getenv("LC_ALL")) == NULL &&
+	    (lang = getenv("LC_CTYPE")) == NULL)
+		/* nothing */;
+
+	language = lang;
+	is_japanese = (lang && strncmp(lang, "ja", 2) == 0);
 }
 
 /*
  * Handle the apropos option.  Cheat by using another program.
  */
-static void
+static int
 do_apropos (char *name) {
-     int len;
-     char *command;
+	char *command;
 
-     len = strlen (getval("APROPOS")) + strlen (name) + 2;
-     command = my_malloc(len);
-     sprintf (command, "%s %s", getval("APROPOS"), name);
-     (void) do_system_command (command, 0);
-     free (command);
+	command = my_xsprintf("'%s' '%Q'", getval("APROPOS"), name);
+	return do_system_command (command, 0);
 }
 
 /*
  * Handle the whatis option.  Cheat by using another program.
  */
-static void
+static int
 do_whatis (char *name) {
-     int len;
-     char *command;
+	char *command;
 
-     len = strlen (getval("WHATIS")) + strlen (name) + 2;
-     command = my_malloc(len);
-     sprintf (command, "%s %s", getval("WHATIS"), name);
-     (void) do_system_command (command, 0);
-     free (command);
+	command = my_xsprintf("'%s' '%Q'", getval("WHATIS"), name);
+	return do_system_command (command, 0);
 }
 
 int
@@ -1014,14 +1226,59 @@ main (int argc, char **argv) {
      char *tmp;
      char *section = 0;
 
+#ifdef __CYGWIN__
+     extern int optind;
+#endif
+
+
+#if 0
+     {
+	/* There are no known cases of buffer overflow caused by
+	   excessively long environment variables. In case you find one,
+	   the simplistic way to fix is to enable this stopgap. */
+	char *s;
+#define CHECK(p,l) s=getenv(p); if(s && strlen(s)>(l)) { fprintf(stderr, "ERROR: Environment variable %s too long!\n", p); exit(1); }
+	CHECK("LANG", 32);
+	CHECK("LANGUAGE", 128);
+	CHECK("LC_MESSAGES", 128);
+	CHECK("MANPAGER", 128);
+	CHECK("MANPL", 128);
+	CHECK("MANROFFSEQ", 128);
+	CHECK("MANSECT", 128);
+	CHECK("MAN_HP_DIREXT", 128);
+	CHECK("PAGER", 128);
+	CHECK("SYSTEM", 64);
+	CHECK("BROWSER", 64);
+	CHECK("HTMLPAGER", 64);
+	/* COLUMNS, LC_ALL, LC_CTYPE, MANPATH, MANWIDTH, MAN_IRIX_CATNAMES,
+	   MAN_ICONV_PATH, MAN_ICONV_OPT, MAN_ICONV_INPUT_CHARSET,
+	   MAN_ICONV_OUTPUT_CHARSET, NLSPATH, PATH */
+     }
+#endif
+
+
 #ifndef __FreeBSD__ 
      /* Slaven Rezif: FreeBSD-2.2-SNAP does not recognize LC_MESSAGES. */
+     setlocale(LC_CTYPE, "");	/* used anywhere? maybe only isdigit()? */
      setlocale(LC_MESSAGES, "");
 #endif
 
+     /* No doubt we'll need some generic language code here later.
+	For the moment only Japanese support. */
+     setlang();
+
      /* Handle /usr/man/man1.Z/name.1 nonsense from HP */
      dohp = getenv("MAN_HP_DIREXT");		/* .Z */
 
+     /* Handle ls.z (instead of ls.1.z) cat page naming from IRIX */
+     if (getenv("MAN_IRIX_CATNAMES"))
+	  do_irix = 1;
+
+     /* Handle lack of ':' in NTFS file names */
+#if defined(_WIN32) || defined(__CYGWIN__)
+     do_win32 = 1;
+#endif
+
      progname = mkprogname (argv[0]);
 
      get_permissions ();
@@ -1049,24 +1306,22 @@ main (int argc, char **argv) {
      while (optind < argc) {
 	  nextarg = argv[optind++];
 
+	  /* is_section correctly accepts 3Xt as section, but also 9wm,
+	     so we should not believe is_section() for the last arg. */
 	  tmp = is_section (nextarg);
-	  if (tmp) {
-	       if (optind < argc) {
-		    section = tmp;
-		    if (debug)
-			 gripe (SECTION, section);
-	       } else {
-		    gripe (NO_NAME_FROM_SECTION, tmp);
-	       }
-	       continue;
+	  if (tmp && optind < argc) {
+		  section = tmp;
+		  if (debug)
+			  gripe (SECTION, section);
+		  continue;
 	  }
 
 	  if (global_apropos)
-	       do_global_apropos (nextarg, section);
+	       status = !do_global_apropos (nextarg, section);
 	  else if (apropos)
-	       do_apropos (nextarg);
+	       status = !do_apropos (nextarg);
 	  else if (whatis)
-	       do_whatis (nextarg);
+	       status = !do_whatis (nextarg);
 	  else {
 	       status = man (nextarg, section);
 
@@ -1078,5 +1333,5 @@ main (int argc, char **argv) {
 	       }
 	  }
      }
-     return !status;
+     return status ? EXIT_SUCCESS : EXIT_FAILURE;
 }
